/* play.c
 *
 * play PCM wav and Ogg Vorbis files over ALSA hardware
 *
 * (C) 2006 Raphael Wimmer <raphael@hcilab.org>
 *
 * using code from http://equalarea.com/paul/alsa-audio.html
 * and from Ogg Vorbis examples
 *
 * licensed under the terms of the GNU GPL v2
 * (http://www.fsf.org/licensing/licenses/gpl.html) 
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <alsa/asoundlib.h>
#include <vorbis/vorbisfile.h>
#include "wav.h"
#include "ogg.h"

#define BUFSIZE 1024 // ALSA frames
#define WAIT 8000 // ALSA timeout in milliseconds. 
                  // Needs to be longer than the buffer content!

#define TYPE_WAV 0
#define TYPE_OGG 1

static char *device = "plughw:0,0";  
snd_pcm_t *playback_handle;
char *buf;
int frame_size = 4; // assume reasonable default :) 

unsigned int sampling_rate, channels, resolution;
snd_pcm_format_t snd_fmt;

int file_type;

// function pointers to format-specific functions
int (*load_file) (const char* file_name) = NULL;
int (*close_file) () = NULL;
int (*get_resolution) (void) = NULL;
snd_pcm_format_t (*get_sound_format) (void) = NULL;
int (*get_sampling_rate) (void) = NULL;
int (*get_channels) (void) = NULL;
int (*fill_buffer) (char* buf, unsigned int bytes) = NULL;

// setup ALSA using code from http://equalarea.com/paul/alsa-audio.html
int init_alsa(int sampling_rate, snd_pcm_format_t snd_fmt, int channels);

int playback_callback (snd_pcm_sframes_t nframes) {
	int err;
	if ((nframes = fill_buffer(buf, nframes * frame_size) / frame_size) <= 0){
		close_file();
		exit(0); // nothing to play
	}
	printf ("playback callback wrote %u frames\n", (unsigned int) nframes);
	
	if ((err = snd_pcm_writei (playback_handle, buf, nframes)) < 0) {
		fprintf (stderr, "write failed (%s)\n", snd_strerror (err));
		snd_pcm_prepare(playback_handle);
	}
	return err;
}

void usage(char* arg0){
	printf("Usage: %s <file.wav|file.ogg>\n", arg0);
}

int main (int argc, char *argv[]) {
	int err;
	snd_pcm_sframes_t frames_to_deliver;
	
	char* file_start;
	FILE* fp;	
	
	if (argc != 2){
		usage(argv[0]);
		exit(0);
	}
	
	/* check file start (first 3 bytes) for file type */
	if ( (fp = fopen(argv[1], "r")) != NULL)
	{
		file_start = calloc(4,1);
		fread(file_start, 3, 1, fp);
		if (ferror(fp) != 0) {
			fclose(fp);
			fprintf(stderr, "Cannot read from file %s!\n", argv[1]); 
			exit(2); // failed to read from file
		}
		fclose(fp);
	} else {
		fprintf(stderr, "Cannot open file%s!\n", argv[1]); 
		exit(1);
	}

	/* determine file type */
	if (strcmp(file_start, "RIF") == 0){
		fprintf(stderr, "WAV file detected.\n");
		file_type = TYPE_WAV;
	} else if (strcmp(file_start, "Ogg") == 0){ 
		fprintf(stderr, "Ogg Vorbis file detected.\n");
		file_type = TYPE_OGG;
	}
	else {
		printf("Unrecognized file format. Exiting.");
		exit(1);
	}
	free(file_start); file_start = NULL;

	/* set up function pointers */
	
	if (file_type == TYPE_WAV){
		load_file = &wav_load_file;
		close_file = &wav_close_file;
		get_sampling_rate = &wav_get_sampling_rate;
		get_sound_format = &wav_get_sound_format;
		get_channels = &wav_get_channels;
		get_resolution = &wav_get_resolution;
		fill_buffer = &wav_fill_buffer;
	} else if (file_type == TYPE_OGG){
		load_file = &ogg_load_file;
		close_file = &ogg_close_file;
		get_sampling_rate = &ogg_get_sampling_rate;
		get_sound_format = &ogg_get_sound_format;
		get_channels = &ogg_get_channels;
		get_resolution = &ogg_get_resolution;
		fill_buffer = &ogg_fill_buffer;
	}

	/* get meta data */
	load_file(argv[1]);
	sampling_rate = get_sampling_rate();
	snd_fmt = get_sound_format();
	channels = get_channels();
	resolution = get_resolution();

	/* << at this point we need to know resolution (snd_fmt),       >>
	 * << channels, and sampling rate of the audio file             >> */
	printf("File: %s, %d Hz, %d bit, %d channels\n", argv[1], sampling_rate, resolution, channels); 
	
	frame_size = channels * resolution / 8;
	printf("Framesize: %d   \n", frame_size);
	// initialize buffer
	buf = malloc(BUFSIZE * frame_size);
	init_alsa(sampling_rate, snd_fmt, channels);
	
	printf("PLAY!\n");
	while (1) {

		/* wait till the interface is ready for data, or 1 second
		 * 			   has elapsed.
		 * 			   			*/
		if ((err = snd_pcm_wait (playback_handle, WAIT)) < 0) {
			fprintf (stderr, "poll failed (%s)\n", strerror (errno));
			break;
		}	           
		/* find out how much space is available for playback data */

		if ((frames_to_deliver = snd_pcm_avail_update (playback_handle)) < 0) {
			if (frames_to_deliver == -EPIPE) {
				fprintf (stderr, "an xrun occured\n");
				break;
			} else {
				fprintf (stderr, "unknown ALSA avail update return value (%u)\n", 
						(unsigned int) frames_to_deliver);
				break;
			}
		}

		frames_to_deliver = frames_to_deliver > BUFSIZE ? BUFSIZE : frames_to_deliver;
		
		/* deliver the data */
		playback_callback (frames_to_deliver);
	}

	snd_pcm_close (playback_handle);
	return(0);
}

int init_alsa(int sampling_rate, snd_pcm_format_t snd_fmt, int channels){
	int err;
	snd_pcm_hw_params_t *hw_params;
	snd_pcm_sw_params_t *sw_params;


	if ((err = snd_pcm_open (&playback_handle, device, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
		fprintf (stderr, "cannot open audio device %s (%s)\n", 
				device,
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_malloc (&hw_params)) < 0) {
		fprintf (stderr, "cannot allocate hardware parameter structure (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_any (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot initialize hardware parameter structure (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_set_access (playback_handle, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
		fprintf (stderr, "cannot set access type (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_set_format (playback_handle, hw_params, snd_fmt)) < 0) {
		fprintf (stderr, "cannot set sample format (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_set_rate_near (playback_handle, hw_params, &sampling_rate, 0)) < 0) {
		fprintf (stderr, "cannot set sample rate (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params_set_channels (playback_handle, hw_params, channels)) < 0) {
		fprintf (stderr, "cannot set channel count (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_hw_params (playback_handle, hw_params)) < 0) {
		fprintf (stderr, "cannot set parameters (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	snd_pcm_hw_params_free (hw_params);

	if ((err = snd_pcm_sw_params_malloc (&sw_params)) < 0) {
		fprintf (stderr, "cannot allocate software parameters structure (%s)\n",
				snd_strerror (err));
		exit (1);
	}
	if ((err = snd_pcm_sw_params_current (playback_handle, sw_params)) < 0) {
		fprintf (stderr, "cannot initialize software parameters structure (%s)\n",
				snd_strerror (err));
		exit (1);
	}
	if ((err = snd_pcm_sw_params_set_avail_min (playback_handle, sw_params, BUFSIZE)) < 0) {
		fprintf (stderr, "cannot set minimum available count (%s)\n",
				snd_strerror (err));
		exit (1);
	}
	if ((err = snd_pcm_sw_params_set_start_threshold (playback_handle, sw_params, 0U)) < 0) {
		fprintf (stderr, "cannot set start mode (%s)\n",
				snd_strerror (err));
		exit (1);
	}
	if ((err = snd_pcm_sw_params (playback_handle, sw_params)) < 0) {
		fprintf (stderr, "cannot set software parameters (%s)\n",
				snd_strerror (err));
		exit (1);
	}

	if ((err = snd_pcm_prepare (playback_handle)) < 0) {
		fprintf (stderr, "cannot prepare audio interface for use (%s)\n",
				snd_strerror (err));
		exit (1);
	}
	 

	return 0;
}
